use std::{iter::Peekable, str::CharIndices};
use unicode_ident::*;

/// Rule        = ident '=' Either? ';' ;
#[derive(Debug, Default)]
pub struct Rule<'a> {
    pub comment: Option<&'a str>,
    pub name: &'a str,
    pub body: RuleKind<'a>,
}

#[derive(Debug, Default)]
pub enum RuleKind<'a> {
    /// Either      = Follow ('|' Follow)* ;
    Either(Vec<RuleKind<'a>>),
    /// Follow      = (Any | Many | Maybe | Not)+ ;
    Follow(Vec<RuleKind<'a>>),
    /// Any         = Not '*' ;
    Any(Box<RuleKind<'a>>),
    /// Many        = Not '+' ;
    Many(Box<RuleKind<'a>>),
    /// Maybe       = Not '?' ;
    Maybe(Box<RuleKind<'a>>),
    /// Not         = '!'? Prime ;
    Not(Box<RuleKind<'a>>),
    /// Group       = '(' Either ')' ;
    Group(Box<RuleKind<'a>>),
    /// ident       = XID_START XID_CONTINUE* ;
    Ident(&'a str),
    /// str         = '"' (!'"' ANY)* '"' ;
    Chr(&'a str),
    /// chr         = "'" (!"'" ANY)* "'" ;
    Str(&'a str),
    /// Nothing
    #[default]
    Empty,
}

impl std::fmt::Display for RuleKind<'_> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            RuleKind::Either(r) => r.iter().enumerate().try_for_each(|(i, r)| {
                if i != 0 {
                    write!(f, " | ")?;
                }
                write!(f, "{r}")
            }),
            RuleKind::Follow(r) => r.iter().enumerate().try_for_each(|(i, r)| {
                if i != 0 {
                    write!(f, " ~ ")?;
                }
                write!(f, "{r}")
            }),
            RuleKind::Any(r) => write!(f, "{r}*"),
            RuleKind::Many(r) => write!(f, "{r}+"),
            RuleKind::Maybe(r) => write!(f, "{r}?"),
            RuleKind::Not(r) => write!(f, "!{r}"),
            RuleKind::Group(r) => write!(f, "({r})"),
            RuleKind::Ident(r) => write!(f, "{r}"),
            RuleKind::Chr(r) => write!(f, "{r:?}"),
            RuleKind::Str(r) => write!(f, "{r:?}"),
            RuleKind::Empty => write!(f, "(!ANY)?"),
        }
    }
}
impl std::fmt::Display for Rule<'_> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let Self {
            comment: _,
            name,
            body,
        } = self;
        // if let Some(comment) = comment {
        //     writeln!(f, "/* {comment} */")?;
        // }
        write!(f, "{name} = {{ {body} }}")
    }
}

pub struct Parser<'a> {
    text: &'a str,
    chars: Peekable<CharIndices<'a>>,
    head: usize,
    tail: usize,
}

impl<'a> Parser<'a> {
    pub fn new(text: &'a str) -> Self {
        Self {
            text,
            chars: text.char_indices().peekable(),
            head: 0,
            tail: 0,
        }
    }
    pub fn error(&mut self, kind: ErrorKind) -> error::Error {
        error::Error::new(self.head, self.tail, kind)
    }
    pub fn start(&mut self) -> &mut Self {
        self.space();
        self.head = self.tail;
        self
    }
    pub fn fragment(&self) -> &'a str {
        let &Self {
            text, head, tail, ..
        } = self;
        &text[head..tail]
    }
    pub fn peek(&mut self) -> Result<char> {
        self.chars
            .peek()
            .map(|(_, c)| *c)
            .ok_or_else(|| self.error(ErrorKind::EndOfInput))
    }
    pub fn take(&mut self) -> Option<(usize, char)> {
        let out = self.chars.next();
        self.tail = match out {
            Some((i, _)) => i + 1,
            None => self.text.len(),
        };
        out
    }
    pub fn take_one(&mut self, f: fn(char) -> bool) -> Result<&mut Self> {
        let Some(&(_, c)) = self.chars.peek() else {
            Err(self.error(ErrorKind::EndOfInput))?
        };
        if f(c) {
            self.take();
            Ok(self)
        } else {
            Err(self.error(ErrorKind::Unexpected(c)))
        }
    }
    pub fn take_many(&mut self, f: fn(char) -> bool) -> &mut Self {
        while self.take_one(f).is_ok() {}
        self
    }
    pub fn space(&mut self) -> &mut Self {
        self.take_many(char::is_whitespace)
    }
}

impl<'a> Parser<'a> {
    pub fn rule(&mut self) -> Result<Rule<'a>> {
        let out = Rule {
            comment: self.comment().ok(),
            name: self.ident()?,
            body: {
                self.space()
                    .take_one(|c| '=' == c)?
                    .either()
                    .unwrap_or_default()
            },
        };
        self.space().take_one(|c| ';' == c).map(|_| out)
    }
    pub fn either(&mut self) -> Result<RuleKind<'a>> {
        let mut out = vec![self.follow()?];
        while self.space().take_one(|c| '|' == c).is_ok() {
            out.push(self.follow()?)
        }
        Ok(match out.len() {
            1 => out.pop().expect("pop should succeed when length is 1"),
            _ => RuleKind::Either(out),
        })
    }
    pub fn follow(&mut self) -> Result<RuleKind<'a>> {
        let mut out = vec![];
        while let Ok(rule) = self.repeat() {
            out.push(rule)
        }
        Ok(match out.len() {
            1 => out.pop().expect("pop should succeed when length is 1"),
            _ => RuleKind::Follow(out),
        })
    }
    pub fn repeat(&mut self) -> Result<RuleKind<'a>> {
        let out = self.not()?;
        let out = match self.space().peek() {
            Ok('*') => RuleKind::Any(out.into()),
            Ok('+') => RuleKind::Many(out.into()),
            Ok('?') => RuleKind::Maybe(out.into()),
            _ => return Ok(out),
        };
        self.take();
        Ok(out)
    }
    pub fn not(&mut self) -> Result<RuleKind<'a>> {
        match self.space().take_one(|c| '!' == c) {
            Ok(_) => Ok(RuleKind::Not(self.prime()?.into())),
            _ => self.prime(),
        }
    }
    pub fn prime(&mut self) -> Result<RuleKind<'a>> {
        Ok(match self.space().peek()? {
            '(' => return self.group(),
            '"' => RuleKind::Str(self.str()?),
            '\'' => RuleKind::Chr(self.chr()?),
            _ => RuleKind::Ident(self.ident()?),
        })
    }
    pub fn group(&mut self) -> Result<RuleKind<'a>> {
        self.take_one(|c| '(' == c)?;
        let out = self.either()?;
        self.take_one(|c| ')' == c)
            .map(|_| RuleKind::Group(out.into()))
    }
    pub fn ident(&mut self) -> Result<&'a str> {
        self.start().take_one(is_xid_start)?;
        self.take_many(is_xid_continue);
        Ok(self.fragment())
    }
    pub fn chr(&mut self) -> Result<&'a str> {
        self.space().take_one(|c| '\'' == c)?;
        self.start().take_many(|c| '\'' != c);
        let out = self.fragment();
        self.take_one(|c| '\'' == c).map(|_| out)
    }
    pub fn str(&mut self) -> Result<&'a str> {
        self.space().take_one(|c| '\"' == c)?;
        self.start().take_many(|c| '\"' != c);
        let out = self.fragment();
        self.take_one(|c| '"' == c).map(|_| out)
    }

    pub fn comment(&mut self) -> Result<&'a str> {
        let start = self.tail;
        while self.space().take_one(|c| '(' == c).is_ok() {
            self.take_one(|c| '*' == c)?;
            loop {
                match self.peek()? {
                    '*' => {
                        self.take_one(|c| '*' == c)?;
                        if self.take_one(|c| ')' == c).is_ok() {
                            break;
                        }
                    }
                    '(' => {
                        let _ = self.comment();
                    }
                    _ => {
                        self.take();
                    }
                }
            }
        }
        let out = &self.text[start..self.tail];
        if out.len() < 2 {
            Ok(out)
        } else {
            Err(self.error(ErrorKind::NoComment))
        }
    }
}

use error::{ErrorKind, Result};
pub mod error {

    pub type Result<T> = std::result::Result<T, Error>;

    #[derive(Debug)]
    pub struct Error {
        place: (usize, usize),
        kind: ErrorKind,
    }
    impl Error {
        pub fn new(start: usize, end: usize, kind: ErrorKind) -> Self {
            Self {
                place: (start, end),
                kind,
            }
        }
        pub fn is_end(&self) -> bool {
            matches!(self.kind, ErrorKind::EndOfInput)
        }
    }
    impl std::error::Error for Error {}
    impl std::fmt::Display for Error {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            let Self {
                place: (start, end),
                kind,
            } = self;
            write!(f, "[{start}..{end}]: {kind} at {end}")
        }
    }

    #[derive(Debug)]
    pub enum ErrorKind {
        Unterminated(Terminable),
        Unexpected(char),
        EndOfInput,
        NoComment,
    }
    impl std::fmt::Display for ErrorKind {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                ErrorKind::Unterminated(t) => write!(f, "unterminated {t:?}"),
                ErrorKind::Unexpected(c) => write!(f, "unexpected character {c}"),
                ErrorKind::EndOfInput => write!(f, "end of input"),
                ErrorKind::NoComment => write!(f, "no comment"),
            }
        }
    }

    #[derive(Debug)]
    pub enum Terminable {
        Comment,
        Group,
        Str,
        Chr,
    }
}
